@@ -1,17 +1,17 @@
-# 
+#
 # This file is part of Config-Model-TkUI
-# 
-# This software is Copyright (c) 2010 by Dominique Dumont.
-# 
+#
+# This software is Copyright (c) 2011 by Dominique Dumont.
+#
 # This is free software, licensed under:
-# 
+#
 #   The GNU Lesser General Public License, Version 2.1, February 1999
-# 
+#
 # copyright at the end of the file in the pod section
 
 package Config::Model::TkUI ;
-BEGIN {
-  $Config::Model::TkUI::VERSION = '1.317';
+{
+  $Config::Model::TkUI::VERSION = '1.329';
 }
 
 use strict;
@@ -30,6 +30,10 @@ use Pod::POM::View::Text ;
 use Tk::Photo ;
 use Tk::PNG ; # required for Tk::Photo to be able to load pngs
 use Tk::DialogBox ;
+use Tk::Adjuster;
+
+use Tk::Pod ;
+use Tk::Pod::Text; # for findpod
 
 use Config::Model::Tk::LeafEditor ;
 use Config::Model::Tk::CheckListEditor ;
@@ -52,19 +56,21 @@ Construct Tk::Widget 'ConfigModelUI';
 
 my $cust_img ;
 my $tool_img ;
+my %gnome_img;
 
 my $mod_file = 'Config/Model/TkUI.pm' ;
 $icon_path = $INC{$mod_file} ;
 $icon_path =~ s/TkUI.pm//;
 $icon_path .= 'Tk/icons/' ;
 
-my $logger = Log::Log4perl::get_logger(__PACKAGE__);
+my $logger = Log::Log4perl::get_logger('TkUI');
 
 no warnings "redefine" ;
 
 sub Tk::Error {
     my ($cw,$error,@locations) = @_;
     my $msg = (ref($error) && $error->can('as_string')) ? $error->as_string : $error ;
+    warn $msg ;
     $msg .= "Tk stack: \n@locations\n";
     $cw -> Dialog ( -title => 'Config::Model error',
                     -text  => $msg,
@@ -92,6 +98,8 @@ sub Populate {
 	$warn_img = $cw->Photo(-file => $icon_path.'dialog-warning.png');
 	# snatched from openclipart-png
 	$tool_img = $cw->Photo(-file => $icon_path.'tools_nicu_buculei_01.png');
+	$gnome_img{next} = $cw->Photo(-file => $icon_path.'gnome-next.png');
+	$gnome_img{previous} = $cw->Photo(-file => $icon_path.'gnome-previous.png');
     }
 
     foreach my $parm (qw/-root/) {
@@ -130,7 +138,8 @@ sub Populate {
 
     my $file_items = [[ qw/command wizard -command/, sub{ $cw->wizard }],
 		      [ qw/command reload -command/, sub{ $cw->reload }],
-		      [ qw/command check  -command/, sub{ $cw->check(1)}],
+		      [ command => 'check for errors',    -command => sub{ $cw->check(1)} ],
+		      [ command => 'check for warnings',  -command => sub{ $cw->check(1,1)} ],
 		      [ qw/command save   -command/, sub{ $cw->save }],
 		      [ command => 'save in dir ...',
                         -command => sub{ $cw->save_in_dir ;} ],
@@ -149,6 +158,7 @@ sub Populate {
 	 # [ qw/command cut   -command/, sub{ $cw->edit_cut }],
 	 [ command => 'copy (Ctrl-C)', '-command', sub{ $cw->edit_copy }],
 	 [ command => 'paste (Ctrl-V)','-command', sub{ $cw->edit_paste }],
+	 [ command => 'find (Ctrl-F)','-command', sub{ $cw->pack_find_widget; }],
 	];
     $menubar->cascade( -label => 'Edit', -menuitems => $edit_items ) ;
 
@@ -190,7 +200,6 @@ sub Populate {
     $cw->{tktree} = $tree ;
 
     # add adjuster
-    require Tk::Adjuster;
     $bottom_frame -> Adjuster()->packAfter($tree, -side => 'left') ;
 
     # add headers
@@ -220,21 +229,26 @@ sub Populate {
     # bind button3 as double-button-1 does not work
     my $b3_sub = sub{my $item = $tree->nearest($tree->pointery - $tree->rooty) ;
 		     $cw->on_select($item)} ;
-    $tree->bind('<Button-3>', $b3_sub) ;
+    $cw->bind('<Button-3>', $b3_sub) ;
+    # pb: rapid click outside the hlist will trigger $b3_sub
+    #$tree->bind('<Double-Button-1>', $b3_sub) ;
 
     # bind button2 to get cut buffer content and try to store cut buffer content
     my $b2_sub = sub{my $item = $tree->nearest($tree->pointery - $tree->rooty) ;
 		     $cw->on_cut_buffer_dump($item)} ;
     $tree->bind('<Button-2>', $b2_sub) ;
 
-    $tree->bind('<Control-c>',  sub{ $cw->edit_copy }) ;
-    $tree->bind('<Control-v>',  sub{ $cw->edit_paste }) ;
+    $cw->bind('<Control-c>',  sub{ $cw->edit_copy }) ;
+    $cw->bind('<Control-v>',  sub{ $cw->edit_paste }) ;
+    $cw->bind('<Control-f>',  sub{ $cw->pack_find_widget }) ;
 
     # bind button2 to get cut buffer content and try to store cut buffer content
     #my $key_sub = sub{my $item = $tree->nearest($tree->pointery - $tree->rooty) ;
     #$cw->on_key_press($item)} ;
     #$tree->bind('<KeyPress>', $key_sub) ;
 
+    my $find_frame = $cw->create_find_widget ;
+
     $args->{-title} = $title;
     $cw->SUPER::Populate($args) ;
 
@@ -256,6 +270,7 @@ sub Populate {
     $cw->Advertise( menubar => $menubar );
     $cw->Advertise( right_frame => $eh_frame );
     $cw->Advertise( ed_frame => $e_frame );
+    $cw->Advertise( find_frame => $find_frame );
 
     $cw->Delegates ;
 }
@@ -290,7 +305,7 @@ sub add_help_menu {
     my $about_sub = sub {
 	$cw->Dialog(-title => 'About',
 		    -text => "Config::Model::TkUI \n"
-		    ."(c) 2008-2010 Dominique Dumont \n"
+		    ."(c) 2008-2011 Dominique Dumont \n"
 		    ."Licensed under LGPLv2\n"
 		   ) -> Show ;
     };
@@ -309,9 +324,20 @@ sub add_help_menu {
 	$db-> Show ;
     };
 
+    my $class = $cw->{root}->config_class_name ;
+    my $man_sub = sub {
+        $cw -> Pod(
+            -tree => 0,
+            -file => "Config::Model::models::".$class ,
+            -title => $class ,
+            -exitbutton => 0,
+        ) ;
+    };
+
     my $help_items = [[ qw/command About -command/, $about_sub ],
 		      [ qw/command Todo  -command/, $todo_sub  ],
 		      [ qw/command Usage -command/, $help_sub  ],
+		      [ command => "$class help", -command => $man_sub ],
 		     ] ;
     $menubar->cascade( -label => 'Help', -menuitems => $help_items ) ;
 }
@@ -347,53 +373,25 @@ sub save_in_dir {
 
 sub check {
     my $cw = shift ;
-    my $show = shift || 0 ;
+    my $show = shift || 0;
+    my $check_warnings = shift || 0;
 
-    # first check for errors, will die on errors
-    eval { $cw->{root}->dump_tree(auto_vivify => 1, full_dump => 1) } ;
+    my $wiz = $cw->setup_wizard(sub{ $cw->check_end($show,@_) ;});
 
-    if ($@) {
-	$cw->handle_error($@) ;
-    }
-    elsif ($show) {
-	$cw->Dialog(-title => 'Check',
-		    -text => "No errors found"
-		   ) -> Show ;
-    }
+    $wiz->start_wizard(experience => $cw->{experience}, stop_on_warning => $check_warnings ) ;
 }
 
-sub handle_error {
-    my $cw = shift;
-    my $e_obj = shift ;
-    my $mode = shift || '' ;
+sub check_end {
+    my $cw = shift ;
+    my $show = shift ;
+    my $has_stopped = shift ;
 
-    my @buttons = qw/ok/ ;
+    $cw->reload if $has_stopped ;
 
-    my $conf_obj = $e_obj->object ;
-    push @buttons, 'edit' if defined $conf_obj ;
-
-    push @buttons, 'trace' unless $mode eq 'trace' ;
-
-    my $d = $cw->DialogBox(-title => 'Error',
-			   -buttons => \@buttons,
-			  ) ;
-
-    if ($mode eq 'trace') {
-	my $t = $d->add('ROText') -> pack;
-	$t->insert(end => $e_obj->trace->as_string);
-    }
-    else {
-	$d->add('Label',
-		-text => $e_obj-> as_string ) -> pack ;
-    }
-
-    my $answer = $d -> Show ;
-
-    if ($answer eq 'trace') {
-	$cw->handle_error($e_obj,$answer) ;
-    }
-    elsif ($answer eq 'edit') {
-	$cw->force_element_display($conf_obj) ;
+    if ($show and not $has_stopped) {
+	$cw->Dialog(-title => 'Check',
+		    -text => "No issue found"
+		   ) -> Show ;
     }
 }
 
@@ -407,18 +405,18 @@ sub save {
     $cw->check() ;
 
     if (defined $cw->{store_sub}) {
-	$logger->info( "Saving data in $trace_dir directory with store call-back" );
-	$cw->{store_sub}->($dir) ;
+       $logger->info( "Saving data in $trace_dir directory with store call-back" );
+       $cw->{store_sub}->($dir) ;
     }
     else {
-	$logger->info( "Saving data in $trace_dir directory with instance write_back" );
-	eval { $cw->{root}->instance->write_back(@wb_args); } ;
-	if ($@) {
-	  $cw -> Dialog ( -title => 'Save error',
-			  -text  => $@->as_string,
-			)
+       $logger->info( "Saving data in $trace_dir directory with instance write_back" );
+       eval { $cw->{root}->instance->write_back(@wb_args); } ;
+       if ($@) {
+         $cw -> Dialog ( -title => 'Save error',
+                         -text  => ref($@) ? $@->as_string : $@ ,
+                       )
             -> Show ;
-	}
+       }
     }
     $cw->{modified_data} = 0 ;
 }
@@ -453,12 +451,12 @@ sub quit {
 
 sub reload {
     my $cw =shift ;
-    my $is_modif          = shift || 0; # whether values where modified
-    my $force_display_obj = shift ;     # force open editor
-    my $path              = shift ;     # force tree to show this path
+    carp "reload: too many parameters" if @_ > 2 ;
+    my $is_modif           = shift || 0; # whether values where modified
+    my $force_display_path = shift ;     # force open editor on this path
 
     $logger->trace("reloading tk tree".
-		   (defined $force_display_obj ? " (forcedisplay)" : '' )
+		   (defined $force_display_path ? " (force display $force_display_path)" : '' )
 		  ) ;
 
     my $tree = $cw->{tktree} ;
@@ -481,8 +479,9 @@ sub reload {
     }
 
     # the first parameter indicates that we are opening the root
-    $sub->(1,$force_display_obj) ;
-    $tree->see($path) if $path and $tree->info(exists => $path);
+    $sub->(1,$force_display_path) ;
+    $tree->see($force_display_path) 
+        if ($force_display_path and $tree->info(exists => $force_display_path));
     $cw->{editor}->reload if defined $cw->{editor};
 }
 
@@ -527,6 +526,13 @@ sub on_cut_buffer_dump {
 	my @keys = ($sel =~ /\n/m) ? split(/\n/,$sel) : ($sel) ;
 	map{ $obj->fetch_with_id($_) } @keys ;
     }
+    elsif ($obj->isa('Config::Model::ListId') and $obj->get_cargo_type !~ /node/ ) {
+	# if array, push values
+	my @v = ($sel =~ /\n/m) ? split(/\n/,$sel) 
+	      : ($sel =~ /,/)   ? split(/,/, $sel)
+	      :                  ($sel) ;
+	$obj->push(@v) ;
+    }
     # else ignore
 
     # display result
@@ -543,7 +549,7 @@ sub force_element_display {
     my $elt_obj = shift ;
 
     $logger->trace( "force display of ".$elt_obj->location );
-    $cw->reload(0, $elt_obj) ;
+    $cw->reload(0, $elt_obj->location) ;
 }
 
 sub prune {
@@ -577,7 +583,7 @@ my %elt_mode = ( leaf => 'none',
 
 sub disp_obj_elt {
     my ($scanner, $data_ref,$node,@element_list) = @_ ;
-    my ($path,$cw,$opening,$fdp_obj) = @$data_ref ;
+    my ($path,$cw,$opening,$fd_path) = @$data_ref ;
     my $tkt = $cw->{tktree} ;
     my $mode = $tkt -> getmode($path) ;
     $logger->trace("disp_obj_elt path $path mode $mode opening $opening "
@@ -616,7 +622,7 @@ sub disp_obj_elt {
 
 	my $elt_loc = $node_loc ? $node_loc.' '.$elt : $elt ;
 
-	$cw->setmode('node',$newpath,$eltmode,$elt_loc,$fdp_obj,$opening,$scan_sub) ;
+	$cw->setmode('node',$newpath,$eltmode,$elt_loc,$fd_path,$opening,$scan_sub) ;
 
 	if ($elt_type eq 'hash') {
 	    $cw->update_hash_image($node->fetch_element($elt), $newpath) ;
@@ -628,7 +634,7 @@ sub disp_obj_elt {
 
 sub disp_hash {
     my ($scanner, $data_ref,$node,$element_name,@idx) = @_ ;
-    my ($path,$cw,$opening,$fdp_obj) = @$data_ref ;
+    my ($path,$cw,$opening,$fd_path) = @$data_ref ;
     my $tkt = $cw->{tktree} ;
     my $mode = $tkt -> getmode($path) ;
     $logger->trace( "disp_hash    path is $path  mode $mode (@idx)" );
@@ -701,10 +707,11 @@ sub disp_hash {
 
 	my $elt_loc = $node_loc ;
 	$elt_loc .=' ' if $elt_loc;
-	$elt_loc .= $element_name.':'.($idx =~ / / ? '"'.$idx.'"' : $idx);
+	# need to keep regexp identical to the one from C::M::Anything:composite_name
+	$elt_loc .= $element_name.':'.($idx =~ /\W/ ? '"'.$idx.'"' : $idx);
 
 	# hide new entry if hash is not yet opened
-	$cw->setmode('hash',$newpath,$eltmode,$elt_loc,$fdp_obj,$opening,$scan_sub) ;
+	$cw->setmode('hash',$newpath,$eltmode,$elt_loc,$fd_path,$opening,$scan_sub) ;
 
 	$prevpath = $newpath ;
     }
@@ -730,17 +737,15 @@ sub update_hash_image {
 }
 
 sub setmode {
-    my ($cw,$type,$newpath,$eltmode,$elt_loc,$fdp_obj,$opening,$scan_sub) = @_ ;
+    my ($cw,$type,$newpath,$eltmode,$elt_loc,$fd_path,$opening,$scan_sub) = @_ ;
     my $tkt = $cw->{tktree} ;
 
-    my $fdp = defined $fdp_obj ? $fdp_obj->location : '';
-
-    my $force_open  = ($fdp and index($fdp,$elt_loc) == 0) ? 1 : 0 ;
-    my $force_match = ($fdp and $fdp eq $elt_loc )         ? 1 : 0;
+    my $force_open  = ($fd_path and index($fd_path,$elt_loc) == 0) ? 1 : 0 ;
+    my $force_match = ($fd_path and $fd_path eq $elt_loc )         ? 1 : 0;
 
     $logger->trace("$type: elt_loc '$elt_loc', opening $opening "
 		   ."eltmode $eltmode force_open $force_open "
-		   . ($fdp ? "on $fdp" : '' ) . "force_match $force_match"
+		   . ($fd_path ? "on '$fd_path' " : '' ) . "force_match $force_match"
 		  ) ;
 
     if ($eltmode ne 'open' or $force_open or $opening ) {
@@ -754,7 +759,7 @@ sub setmode {
 
     # counterintuitive but right: scan will be done when the entry
     # is opened. mode can be open, close, none
-    $scan_sub->($force_open,$fdp_obj) if ( ($eltmode ne 'open') or $force_open) ;
+    $scan_sub->($force_open,$fd_path) if ( ($eltmode ne 'open') or $force_open) ;
 
     if ($force_match) {
 	$tkt->see($newpath);
@@ -777,7 +782,7 @@ sub trim_value {
 
 sub disp_check_list {
     my ($scanner, $data_ref,$node,$element_name,$index, $leaf_object) =@_;
-    my ($path,$cw,$opening,$fdp_obj) = @$data_ref ;
+    my ($path,$cw,$opening,$fd_path) = @$data_ref ;
     $logger->trace( "disp_check_list    path is $path" );
 
     my $value = $leaf_object->fetch ;
@@ -790,7 +795,7 @@ sub disp_check_list {
 
 sub disp_leaf {
     my ($scanner, $data_ref,$node,$element_name,$index, $leaf_object) =@_;
-    my ($path,$cw,$opening,$fdp_obj) = @$data_ref ;
+    my ($path,$cw,$opening,$fd_path) = @$data_ref ;
     $logger->trace( "disp_leaf    path is $path" );
 
     my $std_v = $leaf_object->fetch(qw/mode standard check no silent 1/) ;
@@ -802,7 +807,7 @@ sub disp_leaf {
 	no warnings qw/uninitialized/ ;
 	$img = $cust_img if (defined $value and $std_v ne $value) ;
 	$img = $warn_img if $leaf_object->warning_msg ;
-	$img = $error_img unless $leaf_object->check($value) ;
+	$img = $error_img if $leaf_object->error_msg;
     }
 
     if (defined $img) {
@@ -823,7 +828,7 @@ sub disp_leaf {
 
 sub disp_node {
     my ($scanner, $data_ref,$node,$element_name,$key, $contained_node) = @_;
-    my ($path,$cw,$opening,$fdp_obj) = @$data_ref ;
+    my ($path,$cw,$opening,$fd_path) = @$data_ref ;
     $logger->trace( "disp_node    path is $path" );
     my $curmode = $cw->{tktree}->getmode($path);
     $cw->{tktree}->setmode($path,'open') if $curmode eq 'none';
@@ -842,6 +847,7 @@ sub setup_scanner {
 
        fallback => 'node',
        experience => 'master', #'beginner',
+       check => 'no',
 
        # node callback
        node_content_cb       => \&disp_obj_elt ,
@@ -1026,25 +1032,97 @@ sub edit_paste {
 
 sub wizard {
     my $cw = shift ;
-    my $tree = $cw->{tktree} ;
 
-    # when wizard is run, there's no need to update editor window in
-    # main widget
-    my $wiz = $cw->ConfigModelWizard
-      (
-	-root     => $cw->{root},
-	-store_cb => sub{ $cw->force_element_display(@_)},
-	-end_cb   => sub{ $cw->deiconify; $cw->raise ; },
-       # -show_cb => sub{ $cw->force_element_display(@_)},
-      ) ;
+    my $wiz = $cw->setup_wizard(sub{ $cw->deiconify; $cw->raise ; $cw->reload ;});
 
     # hide main window while wizard is running
     # end_cb callback will raise the main window
     $cw->withdraw ;
 
-    $wiz->start_wizard($cw->{experience}) ;
+    $wiz->prepare_wizard(experience => $cw->{experience}) ;
 }
 
+sub setup_wizard {
+    my $cw = shift ;
+    my $end_sub = shift ;
+
+    # when wizard is run, there's no need to update editor window in
+    # main widget
+    return $cw->ConfigModelWizard
+      (
+	-root     => $cw->{root},
+	-store_cb => sub{ $cw->{modified_data} = 1 ;},
+	-end_cb   => $end_sub,
+      ) ;
+}
+
+sub create_find_widget {
+    my $cw = shift ;
+    my $f = $cw -> Frame(-relief => 'ridge', -borderwidth => 1 , ) ;
+    my $remove_img = $cw->Photo( -file => $icon_path . 'remove.png' );
+
+    $f -> Button (
+        -image   => $remove_img,
+        -command => sub { $f -> packForget() ;},
+        -relief => 'flat',
+    ) -> pack(-side => 'left');
+
+    my $searcher = $cw->{root}->tree_searcher(type => 'all') ;
+    
+    my $search = '';
+    my @result ;
+    $f -> Label (-text => 'Find: ') -> pack(-side => 'left') ;
+    my $e = $f -> Entry (
+        -textvariable => \$search,
+        -validate => 'key',
+        # ditch the search results when find entry is modified.
+        -validatecommand => sub {  @result = () ; return 1 ;} ,
+    ) ->pack (-side => 'left') ;
+    
+    $cw->Advertise( find_entry => $e) ;
+    
+    foreach my $direction (qw/previous next/) {
+        my $s = sub { $cw -> find_item($direction, $searcher,\$search, \@result) ;};
+        $f -> Button (
+            -compound => 'left',
+            -image => $gnome_img{$direction} ,
+            -text   => ucfirst($direction),
+            -command => $s,
+            -relief => 'flat',
+        ) -> pack(-side => 'left');
+    }
+    
+    return $f ;
+}
+
+sub pack_find_widget {
+    my $cw = shift;
+    $cw->Subwidget('find_frame')->pack(-anchor => 'w', -fill => 'x'); 
+    $cw->Subwidget('find_entry') -> focus ;
+}
+
+sub find_item {
+    my ($cw, $direction, $searcher, $search_ref, $result) = @_ ;
+
+    my $find_frame = $cw ->Subwidget('find_frame') ;
+
+    # search the tree, store the result 
+    @$result = $searcher->search($$search_ref) unless @$result ;
+    
+    # and jump in the list widget any time next is hit.
+    if (@$result) {
+        if ( defined $cw->{old_path} and $direction eq 'next' ) {
+            push @$result, shift @$result ;
+        }
+        elsif (defined $cw->{old_path}) {
+            unshift @$result, pop @$result ;
+        }
+        my $path = $result->[0];
+        $cw->{old_path} = $path ;
+
+        $cw->force_element_display($cw->{root}->grab($path)) ;
+    }
+}
 1;
 
 __END__
@@ -1113,12 +1191,40 @@ configuration tree. Beware, there's no "undo" operation.
 
 =item *
 
-Pasting cut buffer into a leaf element will store the content of the
-buffer into the element. This should be used with C<string> or
-C<uniline> elements.
+Pasting cut buffer into:
+
+=over
+
+=item *
+
+a leaf element will store the content of the
+buffer into the element.
+
+=item *
+
+a list element will split the content of the
+buffer with /\n/ or /,/ and push the resulting array at the 
+end of the list element. 
+
+=item *
+
+a hash element will use the content of the cut buffer to create a new key 
+in the hash element. 
 
 =back
 
+=back
+
+=head2 Search
+
+Hit C<Ctrl-F> or use menu C<< Edit -> Search >> to open a search widget at the bottom 
+of the window.
+
+Enter a keyword in the entry widget and click on C<Next> button.
+
+The keyword will be searched in the configuration tree, in element name, in element value and 
+in documentation.
+
 =head2 Editor widget
 
 The right side of the widget is either a viewer or an editor. When
@@ -1138,10 +1244,7 @@ items (mostly missing mandatory values).
 =head1 TODO
 
 - Document widget options. (-root_model and -store_sub, -quit)
-- add better navigation
 - add tabular view ?
-- improve look and feel
-- add search element or search value
 - expand the whole tree at once
 - add plug-in mechanism so that dedicated widget
   can be used for some config Class (Could be handy for
@@ -1153,12 +1256,12 @@ Dominique Dumont, (ddumont at cpan dot org)
 
 =head1 LICENSE
 
-    Copyright (c) 2008-2010 Dominique Dumont.
+    Copyright (c) 2008-2011 Dominique Dumont.
 
     This file is part of Config-Model.
 
     Config-Model is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Lesser Public License as
+    modify it under the terms of the GNU Lesser General Public License as
     published by the Free Software Foundation; either version 2.1 of
     the License, or (at your option) any later version.
 
@@ -1167,7 +1270,7 @@ Dominique Dumont, (ddumont at cpan dot org)
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     Lesser Public License for more details.
 
-    You should have received a copy of the GNU Lesser Public License
+    You should have received a copy of the GNU Lesser General Public License
     along with Config-Model; if not, write to the Free Software
     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
     02110-1301 USA
